import numpy as np
from math import ceil
from convert_to_powers import convert_to_powers
from degree_doubling import degree_doubling
from chebcoeffs import chebcoeffs
from prefs import Prefs
import functools


def chebxroots(func, a = -1, b = 1, method = 'degdub', enhancedoutput = False, prefs = None):
    '''
    Computes the roots of the Chebyshev expansion of a real-valued function
    in a given real interval.

    The function 'func' is approximated by (potentially) piecewise Chebyshev expansions
    .. math::
    	c_{0} + c_{1}  T_{1} (x) + c_{2} T_{2} (x) + ... + c_{N} T_{N} (x)

    These polynomials are then rewritten into  an ordinary series of power form.
    Afterwards a polynomial rootfinder is applied to obtain the roots in the
    interval ('a', 'b').



    Parameters
    ----------
    func : callable
        A real valued function to find the roots of.
    a : float, optional
        Lower boundary of the considered interval. The default is -1.
    b : float, optional
        Lower boundary of the considered interval. The default is 1.
    method : {'convtopow', 'degdub'}, optional
        The method for conversion to ordinary series of power form. Either
        'convtopow'("convert to power") or 'degdub' ("degree doubling") can be
        chosen. The default is 'degdub'.
    enhancedoutput : bool, optional
        If True, borders and coeffs are additionally returned.
        The default is False.

    Returns
    -------
    roots : array_like
        An array containing the roots of the Chebyshev expansion.
    borders : array_like
        An array containing the borders of the subintervals.
    coeffs : list
        A list of the coefficients of the Chebyshev series for each subinterval

    Notes
    -----
    In the case of "degree doubling" one considers a polynomial of degree
    2 * N. Having the coefficients for the Chebychev series :math:'c_j',
    the considered polynomial is
        .. math:: 2 c_0 x^N + \sum_j= 1 c_j (x^{N-j} + x^{N+j})
    The real part of the roots on the unit disk are the roots of the original
    Chebyshev series in ('a', 'b').

    The function is assumed to be analytic, convergence might be slower for a
    non-analytic function.

    The algorithm is taken from [1]_*.

    For the polynomial rootfinding numpy.roots is used.

    References
    ----------
    ..  [1] John P. Boyd "Computing zeros on a real interval through Chebyshev
        expansion and polynomial rootfinding", SIAM Journal on
        Numerical Analysis, vol. 40, no. 5, pp. 1666-1682, 2002

    Examples
    --------
    Compute the root of :math:'\exp(x)-1'
    >>> f = lambda x: numpy.exp(x)-1
    >>> print(chebxroots(f))
    [9.02056208e-16]
    '''
    if prefs == None:
        prefs = Prefs()

    if prefs.memoization == True:
        @functools.lru_cache(maxsize=None)
        def g(x):
            return func(x)
    else:
        def g(x):
            return func(x)

    #set up the maximal degree for splitting
    if method == 'convtopow':
        Nmax = 18
    elif method == 'degdub':
        Nmax = 40
    else:
        print('Please use either convtopow or degdub as method.')
        return


    if prefs.fixeddegree != None:
        N = prefs.fixeddegree
        if isinstance(N, int) == False:
            print('<maxdeg> needs to be either an integer or left empty.')
            return
        if N >= prefs.maxdeg:
            print('Maximal degree is reached, please try a smaller fixed degree or change maxdeg in your Prefs.')
            return
        chebco = chebcoeffs(g,a,b,N)


    #look for an optimal N
    else:
        sz = prefs.stepsize
        for N in range(1,prefs.maxdeg,sz):
            chebco = chebcoeffs(g, a, b, N)
            t = np.sum(np.abs(chebco[ round(2 / 3 * N) : ]))
            if t < prefs.eps:
                break
        #check if N is equal to or larger than maxdeg
        if N == prefs.maxdeg:
            print('Maximal degree is reached, please try a smaller interval.')
            return


    #check if degree is smaller than Nmax a
    if (N <= Nmax) or (prefs.splitting == False) or (prefs.fixeddegree != None):
        if method == 'convtopow':
            roots = convtopowroots(chebco, a, b)
            borders = np.array([a])
            coeffs = [chebco]

        elif method == 'degdub':
            roots = degdubroots(chebco, a, b, prefs)
            borders = np.array([a])
            coeffs = [chebco]


    #else split into equdistant subintervals
    else:
        numintvls = ceil(N / Nmax)

        #check if numintvls is smaller than maxsplit
        if numintvls >= prefs.maxsplit:
            print('Maximal number of subintervals is reached, please try a smaller interval.')

        else:

            subborders = np.linspace(a, b, numintvls+1)

            roots = np.array([])
            coeffs = []
            borders = np.array([])


            for k in range(numintvls):
                intvroots, intvborders, intvcoeffs = chebxroots(g, subborders[k], subborders[k+1], enhancedoutput = True, prefs = prefs)

                roots = np.concatenate((roots, intvroots))
                coeffs += intvcoeffs
                borders = np.concatenate((borders, intvborders))


    #add b to borders
    borders = np.concatenate((borders, np.array([b])))
    #order and remove duplicates
    roots = ordernremove(roots)
    borders = ordernremove(borders)

    if enhancedoutput == True:
        return roots, borders, coeffs
    else:
        return roots


def ordernremove(x):
    """
    Sorts an array and removes duplicates.

    Parameters
    ----------
    x : array_like
        Array to be sorted.

    Returns
    -------
    nodupl : array_like
        The array sorted and without duplicates.

    """
    nodupl = np.unique(x)
    sorted = np.sort(nodupl)
    return sorted

def convtopowroots(chebco, a, b):
    """
    Computes the roots of a Chebyshev polynomial
    with the convert to powers method.

    Parameters
    ----------
    chebco : array_like
        Coefficients of a Chebyshev polynomial
    a : float
        Lower interval boundary.
    b : float
        Upper interval boundary.

    Returns
    -------
    roots : array_like
        Roots of the Chebyshev polynomial.

    """
    polyco = convert_to_powers(chebco)
    roots = np.roots(polyco[::-1])

    #restrict to roots in [a,b]
    roots = roots[np.isreal(roots)]
    roots = (b - a) / 2 * roots + (b + a) / 2

    roots = roots[roots <= b]
    roots = roots[roots >= a]
    roots = np.real(roots)
    return roots

def degdubroots(chebco, a, b, prefs):
    """
    Computes the roots of a Chebyshev polynomial
    with degree doubling.


    Parameters
    ----------
    chebco : array_like
        Coefficients of a Chebyshev polynomial
    a : float
        Lower interval boundary.
    b : float
        Upper interval boundary.
    prefs : TYPE
        DESCRIPTION.

    Returns
    -------
    roots : array_like
        Roots of the Chebyshev polynomial.

    """
    polyco = degree_doubling(chebco)
    roots = np.roots(polyco)

    #restrict to roots on the unit disk, margin to accommodate small errors
    roots = roots[np.absolute(roots) <= 1 + prefs.rounderr]
    roots = roots[np.absolute(roots) >= 1 - prefs.rounderr]
    roots = np.real(roots)
    roots = (b - a) / 2 * roots + (b + a) / 2
    return roots



import numpy as np

#Convert to powers

def convert_to_powers(a):
    '''
    Converts the coefficients of a polynomial written as a Chebyshev series
    into the coefficients in ordinary series of power form.

    Parameters
    ----------
    a : array
        Rank-1 array of coefficients of a Chebyshev series.

    Returns
    -------
    b : array
        An array of the coefficients of the polynomial in ordinary
        series of power form.

    '''
    a_even = a[::2]
    a_odd = a[1::2]


    Q_even = np.identity(len(a_even))
    vector = np.arange(1,len(a_even))

    Q_even_sub = np.diag(2**(2*(vector+1)-3))

    Q_even[1:1+Q_even_sub.shape[0], 1:1+Q_even_sub.shape[1]] = Q_even_sub

    for i in range(len(a_even)):
        for k in range(i):
            Q_even[i-k-1,i] = round(-(2*(i+1)-2*(k+1))*(2*(i+1)-2*(k+1)-1)/(2*(k+1)*(4*(i+1)-2*(k+1)-4))*Q_even[i-k,i],0)

    Q_odd = np.zeros((len(a_odd),(len(a_odd))))

    for i in range(len(a_odd)):
        Q_odd[i,i] = 2**(2*(i+1)-2)
        for k in range(i):
            Q_odd[i-k-1,i] = -((2*(i+1)-2*(k+1)+1)*((i+1)-(k+1)))/((k+1)*(4*(i+1)-2*(k+1)-2))*Q_odd[i-k,i]
    
    b_even = np.dot(Q_even, a_even)
    b_odd = np.dot(Q_odd, a_odd)

    b=np.zeros(len(a))

    b[::2]=b_even
    b[1::2]=b_odd

    return b



import numpy as np

def clenshaw(x,c):
	'''

	Clenshaw scheme for scalar-valued functions



	Parameters
	----------
	x : float
		point at which to evaluate.
	c : array
		weights of the Chebyshev series for a (specific) function.

	Returns
	-------
	y : float
		function value evaluated at x.


    	Notes
    	-----
	C is a column vector, Y = CLENSHAW(X, C) evaluates the Chebyshev
     	expansion
     	Y = P_N(X) = C(1)*T_0(X) + ... + C(N)*T_{N-1}(X) + C(N+1)*T_N(X)
   	using Clenshaw's algorithm.
	 '''
	bk1=0*x
	bk2=0*x
	x=2*x
	n=range(c.size)
	for k in n[c.size:1:-2]:
		bk2=c[k] +x*bk1-bk2
		bk1=c[k-1]+x*bk2-bk1
	if np.mod(c.size-1,2)==1:
		tmp=bk1
		bk1=c[1]+x*bk1-bk2
		bk2=tmp
	y=c[0]+ 0.5*x*bk1-bk2
	return y



from chebxroots import chebxroots
import numpy as np
from chebcoeffs import chebcoeffs
import matplotlib.pyplot as plt
from clenshaw import clenshaw

def plotxroots(f,  x1=-1,x2=1, method = 'degdub',prefs = None, width=15, height=6):
    """
    Returns the plot of the chebyshev approximation of the function f. In addition, if splitting is chosen to be TRUE, the function will eventually (if
    calculated degree is too high) be split into subintervals and then evaluated on each subinterval. The calculated roots are also plotted as well as the
    subinterval borders.

    Parameters
    ----------
    f : callable
        a real valued function to be approximated and to find the roots of.
    x1 : float, optional
        lower boundary of the considered interval. The default is 1.
    x2 : float,optional
        upper boundary of the considered interval. The default is 1.
    method :  {'convtopow', 'degdub'}, optional
        The method for conversion to ordinary series of power form. Either
        'convtopow'("convert to power") or 'degdub' ("degree doubling") can be
        chosen. The default is 'convtopow'.
    prefs : class, optional
        Sets values for some parameters of chebxroots. See also ../../chebxroots/prefs. The default is None.
    width: float, optional
        width of plot
    height: float, optioanol
        height of plot

    Returns
    -------
    ax : Axes Instance of matplotlib.pyplot containing axis, Tick, Lin2D etc.

    Examples
    --------
    Plot approximated graph and roots of :math:'\sin(x)'
    >>> f = lambda x: numpy.´sin(x)
    >>> ax= plotxroots(f,x1=0,x2=100)

    """
    #create plot
    fig,ax=plt.subplots()
    fig.set_figwidth(width)
    fig.set_figheight(height)
    #calculate roots, subinterval borders, and coefficients (for each subinterval)
    roots_f,borders_f,coefs_f=chebxroots(f, x1, x2, method=method, prefs =prefs, enhancedoutput=True)
    yvals=[]
    xvals=[]
    #evaluate function through clenshaw algorithm in every subinterval
    for j in range(len(coefs_f)):
        #subinterval borders
        b1=borders_f[j]
        b2=borders_f[j+1]
        x=np.linspace(b1,b2,round((b2-b1)*100))
        #chebyshev coefficients for the subinterval
        c=chebcoeffs(f,a=b1,b=b2,deg=len(coefs_f[j]))
        #evaluate function on subinterval
        y=clenshaw((2*x-(b1+b2))/(b2-b1),c)
        yvals.append(y)
        xvals.append(x)
        #draw a vertical line to visualize borders of subintervals
        ax.axvline(b2, color='r', linestyle='--')
        roots_interval=roots_f[roots_f<=b2]
        roots_interval=roots_interval[roots_interval>=b1]
        #scatter plot of calculated roots
        ax.scatter(roots_interval,np.zeros(len(roots_interval)), label= 'roots found in interval {}'.format(j))

    xvals=np.concatenate(xvals).ravel()
    yvals=np.concatenate(yvals).ravel()
    ax.plot(xvals,yvals)
    ax.grid(True, which='both', linestyle= '--')
    plt.legend(loc='upper left')
    plt.xticks(borders_f)
    plt.title('Chebyshev approximation of the function f, approximated on {} intervals'.format(len(coefs_f)))
    plt.legend(loc='upper left')
    return ax



class Prefs:
    '''
    Sets values for some parameters of chebxroots.

    Parameters
    ----------
    fixeddegree : int
        Fix the degree of the Chebyshev expansion. If left empty, an optimal
        degree is determined.
        The default is None.
    maxdeg : int
        Maximal degree determined with stopping criterion sum_{j=[2/3 N]}^N |a_j| < eps
        from [1] where the a_j are Chebychev coefficients.  The default is 500.
    eps : float
        Epsilon for stopping criterion from [1]. The default is 1e-11.
    maxsplit : int
        Maximal number of subintervals. The default is 160.
    rounderr : float
        Rounding error used in degdubroots to restrict to roots on unit disk. The default is 0.000001.
    splitting : bool
        If True, it is allowed to split the considered interval in subintervals.
        The default is True.
    memoization : bool
        If true, memoization is used for the evaluation of the function in chebsroots.
    stepsize : int
        Stepsize for finding the optimal degree. The default is 1.

    References
    ----------
    ..  [1] John P. Boyd "Computing zeros on a real interval through Chebyshev
        expansion and polynomial rootfinding", SIAM Journal on
        Numerical Analysis, vol. 40, no. 5, pp. 1666-1682, 2002
    '''

    def __init__(self, fixeddegree = None, maxdeg = 500, eps = 1e-3, maxsplit = 160, rounderr=0.000001, splitting = True, memoization = False, stepsize = 1):
        self.fixeddegree = fixeddegree
        self.maxdeg = maxdeg
        self.maxsplit = maxsplit
        self.rounderr = rounderr
        self.eps = eps
        self.splitting = splitting
        self.memoization = memoization
        self.stepsize = 1



import numpy as np
import functools


def chebcoeffs(f , a = -1, b = 1  , deg = 17):
    '''
    Computes the coefficients for the Chebyshev expansion of
    a fucntion in a given interval.

    Parameters
    ----------
    func : callable, f(x, *args)
    	A function that takes a float as argument and returns a float.
    a : float, optional
    	Lower boundary of the considered interval. The default is -1.
    b : float, optional
    	Lower boundary of the considered interval. The default is 1.
    deg : int, optional
    	Degree of the Chebyshev expansion. The default is 17.
        
    Returns
    -------
    coeffs : array
    	An array of the coefficients of the Chebyshev expansion.

    '''
    N = deg

    # create chebyshev points
    xk = (b - a) / 2 * chebpts(N+1) + (b + a) / 2

    #compute interpolation matrix
    p = np.ones(N+1)
    p[0] = 2
    p[N] = 2
    p = p[np.newaxis]
    J = 2 / (p.T @ p * N)

    i = np.arange(N+1)[np.newaxis]
    I = np.cos(np.pi * (i.T @ i) / N)

    I = I * J

    #compute the values of f at the Chebyshev points. vectoriz f just to be safe
    fvec = np.vectorize(f)
    fvals = fvec(xk)

    #compute the weights

    coeffs = I.dot(fvals)
    return coeffs

@functools.lru_cache(maxsize=200)
def chebpts(n):
    '''
    Yields Chebyshev points of the second kind.

    Parameters
    ----------
    n : int
        The amount of Chebyshev points of the second kind.

    Returns
    -------
    x : array
        An array containing the Chebyshev points of the second kind.

    '''

    # Special case (no points)
    if n == 0:
        x = []
    # Special case (single point)
    elif n == 1:
        x = 0

    else:
        # Chebyshev points:
        m = n - 1

        #x = np.sin(np.pi*(np.r_[-m:m+1:2]/(2*m))) use for chebfun
        x = np.cos(np.pi*(np.r_[0:m+1]/m))


    return x



import numpy as np
def degree_doubling(a):
    '''
    Given coefficients of a Chebyshev series, yields the coefficients
    used for the degree doubling method.


    Parameters
    ----------
    a : array_like
        Coefficients of a Chebyshev series.

    Returns
    -------
    b : array
        Coefficients of the polynomial obtained by degree doubling.

    '''
    n = len(a)-1
    b = np.zeros(2*n+1)
    b[:n+1] += np.flip(a)
    b[n:] += a
    return b



import setuptools
def readme():
    with open("README.md") as f:
        return f.read()

setuptools.setup(
  name='chebxroots',
  version='0.1.1',
  author='Anna Clara Wendler, Mira Kristin Juergens, Lars Lammers',
  author_email = 'lars.lammers@stud.uni-goettingen.de' ,
  packages=["chebxroots"],
  scripts=['bin/script1','bin/script2'],
  url='https://gitlab.gwdg.de/annaclara.wendler/scientific-computing-chebychev-roots/',
  license='MIT',
  description='A package for computing roots of generic smooth functions.',
  #long_description=open('README.txt').read(),
  install_requires=[
       "scipy",
       "numpy",
       "pytest",
       "matplotlib"
   ],
)
